#ifndef SVM02_H
#define SVM02_H

#include "processor.h"
#include "vm_exceptions.h"

///@class SVM02
///@brief Implements the SVM 0.2 processor's instruction set
class SVM02 : public Processor
{
public:
	SVM02(VMState& state):Processor(state){}

public:
	///@method handleInstruction
	///@brief Instruction handling for SVM 0.2
	virtual void handleInstruction(int32_t icode);
	
	///@method checkStackSize
	///@brief Make sure the stack has at least the given number of elements
	///@param min_size The minimal number of elements on stack
	///@throw CorruptStackException if the stack contains less elements than necessary
	void checkStackSize(int min_size)
	{
		if(m_state.sp < min_size)
		{
			throw CorruptStackException();
		}
	}
	
	///@method checkStackFreeSpace
	///@brief Make sure that there is enough space to insert a given number of elemnts on stack
	///@param min_free_space The desired number of elements
	///@throw StackOverflowException if there's not enough space on stack to hold the desired number of elements
	void checkStackFreeSpace(int min_free_space)
	{
		if(m_state.stack.size() - m_state.sp < min_free_space)
		{
			throw StackOverflowException();
		}
	}
	
	///@method checkMemoryAddress
	///@brief Checks if the given address is within the RAM bounds
	///@param addr The address to be checked
	///@throw InvalidMemoryAddressException if the given address is invalid
	void checkMemoryAddress(int32_t addr)
	{
		if(addr < 0 || addr >= m_state.memory.size())
		{
			throw InvalidMemoryAddressException();
		}
	}
	
	///@method checkProgramMemoryAddress
	///@brief Checks if the given address is within the program's memory bounds
	///@param addr The address to be checked
	///@throw InvalidMemoryAddressException if the given address is invalid
	void checkProgramMemoryAddress(int32_t addr)
	{
		if(addr < 0 || addr >= m_state.program_memory.size())
		{
			throw InvalidMemoryAddressException();			
		}
	}
	
private:
	//Input-output handling
	void in();
	void out();
	
	//ALU
	void add();
	void sub();
	void mul();
	void div();
	void mod();
	void neg();
	void inc();
	void dec();
	void and_(); //and, or, not, xor are keywords
	void or_();
	void not_();
	void xor_();
	void shr();
	void shl();
	
	//Stack manipulation
	void pop();
	void dup();
	void swp();
	void ovr();
	
	//Memory manipultion
	void load();
	void stor();
	
	//Jumps
	void jmp();
	void je();
	void jne();
	void jg();
	void jge();
	void jl();
	void jle();
	
	//Misc
	void nop();
	void halt();
	
	void rf();
	void crf();
};
#endif //SVM02_H
