#ifndef VIRTUAL_MACHINE_H
#define VIRTUAL_MACHINE_H

#include "processor.h"
#include "debugger.h"
#include "vm_state.h"
#include "vm_exceptions.h"
#include "logger.h"
#include <vector>
#include <fstream>
#include <string>
#include <memory>
#include <stdexcept>

///@class VirtualMachine
///@brief Implements SVM architecture
template <class ProcessorType, class DebuggerType>
class VirtualMachine
{
public:
	explicit VirtualMachine(unsigned stack_size = DefaultStackSize, 
				unsigned mem_size = DefaultMemorySize): m_processor(new ProcessorType(m_state)), m_debugger(new DebuggerType(m_state))
	{
		LOG_INF("Created new VM instance");
		LOG_INF("Initializing VM state");
		m_state.stack.resize(stack_size);
		m_state.memory.resize(mem_size);
		m_state.ip = 0;
		m_state.sp = 0;
		m_state.hf = false;
		m_state.rf = false;
		LOG_INF("Complete");
	}
	
public:
	///@method load
	///@brief Load the program into memory from the specified filer
	void load(const char* file_name)
	{
		std::ifstream in(file_name, std::ios::binary);
		LOG_INF("Loading program from file [" << file_name << "]");
		if(!in.is_open())
		{
			LOG_ERR("Failed to open [" << file_name << "]");
			throw std::runtime_error("failed to open input file");
		}
		int32_t instruction;
		while(in.read(reinterpret_cast<char*>(&instruction), sizeof(int32_t)))
		{
			m_state.program_memory.push_back(instruction);
		}
		LOG_INF("Program loaded into memory");
	}
	
	///@method run
	///@brief Start program execution
	void run()
	{
			LOG_INF("Starting program execution");
			try
			{
				while(m_state.ip < m_state.program_memory.size() && !m_state.hf)
				{
					tick();
				}
				LOG_INF("Completed program execution");
			}
			catch(const CorruptStackException&)
			{
				LOG_ERR("Program error: corrupt stack.");
			}
			catch(const StackOverflowException&)
			{
				LOG_ERR("Program error: stack overflow.");
			}
			catch(const InvalidMemoryAddressException&)
			{
				LOG_ERR("Program error: invalid memory address.");
			}
			catch(const UnknownInstructionException&)
			{
				LOG_ERR("Program error: unknown instruction.");
			}
			catch(const VMException&)
			{
				LOG_ERR("Program error: unknown error.");
			}
			catch(const std::runtime_error& e)
			{
				LOG_ERR("Runtime error occured: [" << e.what() << "]");
			}
	}

public:
	static const int DefaultStackSize = 1024;
	static const int DefaultMemorySize = 1024;
	
private:
	///@method tick
	///@brief Implement one tick
	void tick()
	{
		m_debugger->dbg();
		m_processor->handleInstruction(m_state.program_memory[m_state.ip]);
		m_state.ip++;
	}
	
private:
	///@brief Stack, memory, stack and instruction pointers
	VMState m_state;
	
	///@brief The instruction processor
	std::auto_ptr<Processor> m_processor;
	
	///@brief Debugger
	std::auto_ptr<Debugger> m_debugger;
};
#endif //VIRTUAL_MACHINE_H
