#include "code_generator.h"
#include "logger.h"
#include <stdexcept>
#include <cstdlib>
#include <algorithm>

CodeGenerator::CodeGenerator()
{
	initInstructionTable();
}

void CodeGenerator::initInstructionTable()
{
	m_instr["IN"] = -1;
	m_instr["OUT"] = -2;
	m_instr["ADD"] = -3;
	m_instr["SUB"] = -4;
	m_instr["MUL"] = -5;
	m_instr["DIV"] = -6;
	m_instr["MOD"] = -7;
	m_instr["NEG"] = -8;
	m_instr["INC"] = -9;
	m_instr["DEC"] = -10;
	m_instr["AND"] = -11;
	m_instr["OR"] = -12;
	m_instr["NOT"] = -13;
	m_instr["XOR"] = -14;
	m_instr["SHL"] = -15;
	m_instr["SHR"] = -16;
	m_instr["POP"] = -17;
	m_instr["DUP"] = -18;
	m_instr["SWP"] = -19;
	m_instr["OVR"] = -20;
	m_instr["LOAD"] = -21;
	m_instr["STOR"] = -22;
	m_instr["JMP"] = -23;
	m_instr["JE"] = -24;
	m_instr["JNE"] = -25;
	m_instr["JG"] = -26;
	m_instr["JGE"] = -27;
	m_instr["JL"] = -28;
	m_instr["JLE"] = -29;
	m_instr["NOP"] = -30;
	m_instr["HALT"] = -31;
	m_instr["RF"] = -101;
	m_instr["CRF"] = -102;
}

void CodeGenerator::generateInstruction(const std::string& str, CompilerState state)
{
	switch(state)
	{
	case Comment:
		break;
		
	case Instruction:
		generateSimpleInstructionCode(str);
		break;
		
	case LabelDef:
		generateLabelDefCode(str);
		break;
		
	case LabelRef:
		generateLabelRefCode(str);
		break;
			
	case CharConst:
		generateCharacterConstantCode(str);
		break;
		
	case ImmediateValue:
		generateImmediateValueCode(str);
		break;
		
	case StrConst:
		generateStringConstantCode(str);
		break;
	}
}

void CodeGenerator::generateSimpleInstructionCode(const std::string& s)
{
	std::string str = s;
	convertToUppercase(str);
	if(m_instr.count(str))
	{
		m_code.push_back(m_instr[str]);
	}
	else if(str == "PRINT")
	{
		generatePRINTCode();
	}
	else
	{
		LOG_ERR("Unknown instruction '" << str<< "'");
		throw std::runtime_error("");
	}
}

void CodeGenerator::generatePRINTCode()
{
	m_code.push_back(m_instr["RF"]);
	int32_t prn = m_code.size();
	int32_t end = prn + 7;
	m_code.push_back(m_instr["DUP"]);
	m_code.push_back(0);
	m_code.push_back(end);
	m_code.push_back(m_instr["JE"]);
	m_code.push_back(m_instr["OUT"]);
	m_code.push_back(prn);
	m_code.push_back(m_instr["JMP"]);
	m_code.push_back(m_instr["CRF"]);
	m_code.push_back(m_instr["POP"]);
}

void CodeGenerator::convertToUppercase(std::string& s)
{
   std::transform(s.begin(), s.end(), s.begin(), toupper);
}

void CodeGenerator::generateLabelDefCode(const std::string& str)
{
	if(m_labels.count(str) && m_labels[str].address != -1)
	{
		LOG_ERR("Duplicate label '" << str << "'");
		throw std::runtime_error("");
	}
	m_labels[str].address = m_code.size();
}

void CodeGenerator::generateLabelRefCode(const std::string& str)
{
	m_code.push_back(0); //Insert label address placeholder
	m_labels[str].ref.push_back(m_code.size() - 1);
}

void CodeGenerator::generateImmediateValueCode(const std::string& str)
{
	int32_t v = atol(str.c_str());
	m_code.push_back(abs(v));
	
	//Instruction codes are negative, so we need
	//two steps to place a negative value on the
	//stack: push the absolute value of our integer
	//then execute the NEG command
	if(v < 0) 
	{
		m_code.push_back(m_instr["NEG"]);
	}
}

void CodeGenerator::generateCharacterConstantCode(const std::string& str)
{
	m_code.push_back(static_cast<int32_t>(str[0]));
}

void CodeGenerator::generateStringConstantCode(const std::string& str)
{
	for(int i = str.size()-1; i >= 0; --i)
	{
		m_code.push_back(static_cast<int32_t>(str[i]));
	}
}

void CodeGenerator::finalizeCode()
{
	//Insert correct addresses of labels.
	for(LabelTable::iterator li = m_labels.begin();
		li != m_labels.end();
		++li)
	{
		if(li->second.address == -1)
		{
				LOG_ERR("Undefined label '" << li->first << "'");
				throw std::runtime_error("");
		}
		else
		{
			for(int i = 0; i < li->second.ref.size(); ++i)
			{
				m_code[li->second.ref[i]] = li->second.address;
			}
		}
	}
}

void CodeGenerator::dumpCode(std::ostream& out)
{
	finalizeCode();
	m_code.push_back(m_instr["NOP"]); //Special case when a label is at the very end of the file
	
	//Write code to output stream
	LOG_INF("Writing output");
	for(int i = 0; i < m_code.size(); ++i)
	{
		out.write(reinterpret_cast<char*>(&m_code[i]), sizeof(int32_t));
	}
	LOG_INF("Complete, " << sizeof(int32_t)*m_code.size() << " bytes written to output.");
	
	m_code.clear();
	m_labels.clear();
}

