///@file compiler.cpp
///@author grigoryj
#include "compiler.h"
#include "logger.h"
#include <fstream>
#include <cctype>
#include <cstdlib>
#include <stdexcept>
	
Compiler::Compiler():m_state(NewLine){}

void Compiler::processInput(std::istream& in)
{
		while(in.read(&m_next_char, 1))
		{
			if(returnToInitialState()) continue;
			switch(m_state)
			{
			case NewLine:
					parseNewLine();
					break;
					
			case LabelDef:
			case LabelRef:
				parseLabel();
				break;
				
			case ImmediateValue:
				parseImmediateValue();
				break;
				
			case Instruction:
				parseInstruction();
				break;
				
			case Comment:
				parseComment();
				break;
				
			case CharConst:
				parseCharacterConstant();
				break;
				
			case StrConst:
				parseStringConstant();
				break;
			}
		}
		m_codeGen.generateInstruction(m_line, m_state); //There might remain untranslated input
		
}

bool Compiler::returnToInitialState()
{
	if(isspace(m_next_char) && m_state != NewLine && 
			   m_state != Comment && m_state != StrConst) //whitespace is allowed in comments and string constants
	{
		m_codeGen.generateInstruction(m_line, m_state); //Generate code for the current instruction
		m_line = "";
		m_state = NewLine;
		return true;
	}
	return false;
}

void Compiler::dumpCode(std::ostream& out)
{
		
	m_codeGen.dumpCode(out);
	
	//Reset state
	m_state = NewLine;
	m_line = "";
}


void Compiler::parseNewLine()
{
	//Decide which state we must go to
	if(m_next_char == ':') m_state = LabelDef;
	else if(m_next_char == '@') m_state = LabelRef;
	else if(m_next_char == '\'') m_state = CharConst;
	else if(m_next_char == '"') m_state = StrConst;
	else if(m_next_char == ';') m_state = Comment;
	else if(m_next_char == '-' || isdigit(m_next_char))
	{
		m_state = ImmediateValue;
		m_line.push_back(m_next_char);
	}
	else if(isalpha(m_next_char))
	{
		m_state = Instruction;
		m_line.push_back(m_next_char);
	}
	else if(isspace(m_next_char));
	else
	{
		LOG_ERR("Invalid character '" << m_next_char<< "' in input file");
		throw std::runtime_error("");
	}
}

void Compiler::parseLabel()
{
	if(isalnum(m_next_char) || m_next_char == '_') 
	{
		m_line.push_back(m_next_char);
	}
	else
	{
		LOG_ERR("Invalid character '" << m_next_char << "' in input file");
		throw std::runtime_error("");
	}
}

void Compiler::parseImmediateValue()
{
	if(isdigit(m_next_char))
	{
		m_line.push_back(m_next_char);
	}
	else
	{
		LOG_ERR("Invalid character '" << m_next_char << "' in input file");
		throw std::runtime_error("");
	}
}

void Compiler::parseInstruction()
{
	if(isalpha(m_next_char))
	{
		m_line.push_back(m_next_char);
	}
	else
	{
		LOG_ERR("Invalid character '" << m_next_char << "' in input file");
		throw std::runtime_error("");
	}
}

void Compiler::parseComment()
{
	//All comments end at newline
	if(m_next_char == '\n') m_state = NewLine;
}

void Compiler::parseCharacterConstant()
{
	static bool cc_slash = false;
	if(m_line.empty())
	{
		if(m_next_char == '\\')
		{
			cc_slash = true;
		}
		else 
		{
			if(cc_slash)
			{
				m_line.push_back(getEscapeChar(m_next_char));
				cc_slash = false;
			}
			else
			{
				m_line.push_back(m_next_char);
			}
		}
	}
	else
	{
			if(m_next_char == '\'' && m_line.length() == 1);
			else
			{
					LOG_ERR("Invalid character constant");
				throw std::runtime_error("");
			}
	}
}

void Compiler::parseStringConstant()
{
	static bool cc_slash = false;
	if(m_next_char != '"')
	{
		if(m_next_char == '\\')
		{
			cc_slash = true;
		}
		else
		{
			if(cc_slash)
			{
				m_line.push_back(getEscapeChar(m_next_char));
				cc_slash = false;
			}
			else
			{
				m_line.push_back(m_next_char);
			}
		}
	}
	else
	{
		m_line.push_back(0);
		m_codeGen.generateInstruction(m_line, m_state);
		m_line = "";
		m_state = NewLine;
	}
}

char Compiler::getEscapeChar(char c)
{
	switch(c)
	{
	case 'n':
		return ('\n');
		break;
								
	case 't':
		return ('\t');
		break;
							
	case 'a':
		return ('\a');
		break;
							
	case '\\':
		return ('\\');
		break;
								
	case '0':
		return (0);
		break;
		
	default:
		return (c); //This will also handle single and double quotes
	}
}
